@@ -30,6 +30,7 @@ from ..helpers import popen_with_error_handling
 from ..helpers import dash_open
 from ..helpers import iter_separated
 from ..helpers import eval_escapes
+from ..platform import is_win32, swidth
 
 from . import BaseTestCase, FakeInputs
 
@@ -348,19 +349,23 @@ def test_chunkerparams():
     assert ChunkerParams('fixed,4096') == ('fixed', 4096, 0)
     assert ChunkerParams('fixed,4096,200') == ('fixed', 4096, 200)
     # invalid values checking
-    with pytest.raises(ValueError):
+    borg2 = False  # for borg < 2, we only emit a warning, but do not raise ArgumentTypeError for some cases
+    with pytest.raises(ArgumentTypeError):
         ChunkerParams('crap,1,2,3,4')  # invalid algo
-    with pytest.raises(ValueError):
-        ChunkerParams('buzhash,5,7,6,4095')  # too small min. size
-    with pytest.raises(ValueError):
+    if borg2:
+        with pytest.raises(ArgumentTypeError):
+            ChunkerParams('buzhash,5,7,6,4095')  # too small min. size
+    with pytest.raises(ArgumentTypeError):
         ChunkerParams('buzhash,19,24,21,4095')  # too big max. size
-    with pytest.raises(ValueError):
-        ChunkerParams('buzhash,23,19,21,4095')  # violates min <= mask <= max
-    with pytest.raises(ValueError):
-        ChunkerParams('fixed,63')  # too small block size
-    with pytest.raises(ValueError):
+    if borg2:
+        with pytest.raises(ArgumentTypeError):
+            ChunkerParams('buzhash,23,19,21,4095')  # violates min <= mask <= max
+    if borg2:
+        with pytest.raises(ArgumentTypeError):
+            ChunkerParams('fixed,63')  # too small block size
+    with pytest.raises(ArgumentTypeError):
         ChunkerParams('fixed,%d,%d' % (MAX_DATA_SIZE + 1, 4096))  # too big block size
-    with pytest.raises(ValueError):
+    with pytest.raises(ArgumentTypeError):
         ChunkerParams('fixed,%d,%d' % (4096, MAX_DATA_SIZE + 1))  # too big header size
 
 
@@ -896,6 +901,29 @@ def test_progress_percentage_sameline(capfd, monkeypatch):
     assert err == ' ' * 4 + '\r'
 
 
+@pytest.mark.skipif(is_win32, reason='no working swidth() implementation on this platform')
+def test_progress_percentage_widechars(capfd, monkeypatch):
+    st = 'スター・トレック'  # 'startrek' :-)
+    assert swidth(st) == 16
+    path = '/カーク船長です。'  # 'Captain Kirk'
+    assert swidth(path) == 17
+    spaces = ' ' * 4  # to avoid usage of '...'
+    width = len('100%') + 1 + swidth(st) + 1 + swidth(path) + swidth(spaces)
+    monkeypatch.setenv('COLUMNS', str(width))
+    monkeypatch.setenv('LINES', '1')
+    pi = ProgressIndicatorPercent(100, step=5, start=0, msg=f'%3.0f%% {st} %s')
+    pi.logger.setLevel('INFO')
+    pi.show(0, info=[path])
+    out, err = capfd.readouterr()
+    assert err == f'  0% {st} {path}{spaces}\r'
+    pi.show(100, info=[path])
+    out, err = capfd.readouterr()
+    assert err == f'100% {st} {path}{spaces}\r'
+    pi.finish()
+    out, err = capfd.readouterr()
+    assert err == ' ' * width + '\r'
+
+
 def test_progress_percentage_step(capfd, monkeypatch):
     # run the test as if it was in a 4x1 terminal
     monkeypatch.setenv('COLUMNS', '4')
